import rclpy
import os
from modern_robotics import FKinBody
from ament_index_python.packages import get_package_share_directory
from mr_urdf_loader import loadURDF
from rclpy.node import Node
from math import sin, cos, pi
import numpy as np
from std_msgs.msg import Float64
from nav_msgs.msg import Path
from geometry_msgs.msg import PoseStamped
from geometry_msgs.msg import Pose
from visualization_msgs.msg import Marker
import transformations
from scipy.spatial.transform import Rotation
import time

class Forward_Kinematics_Node_1Dof(Node):

    def __init__(self):
        super().__init__('forward_kinematics_node_1dof')
        # Create a subscriber, initialising the transformation
        self.subscription = self.create_subscription(
            Float64,
            '/dof1/reference_angles',
            self.listener_callback,
            10)
        
        urdf_file_path = os.path.join(
            get_package_share_directory('iiwa_description'), 'urdf', 'iiwa14.urdf')
        self.robot_model = loadURDF(urdf_file_path)
        self.M = self.robot_model["M"]
        self.Slist = self.robot_model["Slist"]
        self.Mlist = self.robot_model["Mlist"]
        self.Glist = self.robot_model["Glist"]
        self.Blist = self.robot_model["Blist"]

        urdf_file_path_simple_robot = os.path.join(
            get_package_share_directory('simple_robot_1dof'), 'urdf', 'simple_robot.urdf')
        self.simple_robot_model = loadURDF(urdf_file_path_simple_robot)
        self.simple_robot_M = self.simple_robot_model["M"]
        self.simple_robot_Slist = self.simple_robot_model["Slist"]
        self.simple_robot_Mlist = self.simple_robot_model["Mlist"]
        self.simple_robot_Glist = self.simple_robot_model["Glist"]
        self.simple_robot_Blist = self.simple_robot_model["Blist"]


        # create two publishers for the RViz simulation - one for the whole path and one for the marker
        self.path_publisher_ = self.create_publisher(Path, '/simple_robot/visualization_path', 10)
        self.point_publisher_ = self.create_publisher(Marker, '/simple_robot/visualization_marker', 10)
        # create a publisher for the plotjuggler angles
        self.angle_publisher_ = self.create_publisher(Float64, '/simple_robot/angles_1dof/plotjuggler', 10)
        # create a publisher for the pose
        self.pose_publisher_simple_robot_ = self.create_publisher(Pose, '/simple_robot/target_pose', 10)
        self.pose_publisher_kuka_robot_ = self.create_publisher(Pose, '/kuka_iiwa/target_pose', 10)
        #RViz
        self.path = Path()
        self.path.header.frame_id = "map"
        self.path.header.stamp = self.get_clock().now().to_msg()

        self.index = 0

    def listener_callback(self, msg):
        # Get the message
        self.phi = msg.data
        # call the transformation function
        if self.phi == 0:
            self.index += 1

        self.transformation_kuka()
        self.transformation_simple_robot()
        
    def transformation_kuka(self):
        # Define the desired rotation here
        theta = - pi / 1.8
        # Split theta
        theta_fraction = self.phi * theta

        self.publish_angle(theta_fraction)

        # Define the matrix
        intermediate_rotation = np.array([
            [cos(theta_fraction), -sin(theta_fraction), 0, 0],
            [sin(theta_fraction), cos(theta_fraction), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ])

        self.joint_angles = np.zeros(len(self.Slist[0]))
        self.joint_angles[0] = np.pi
        self.joint_angles[1] = np.pi/2
        self.joint_angles[3] = np.pi/2
        self.joint_angles[5] = np.pi/2

        start_forward_kinematics = FKinBody(self.M, self.Blist, self.joint_angles)

        start_position = start_forward_kinematics[:4 ,3]
        start_orientation = Rotation.from_matrix(start_forward_kinematics[:3, :3]).as_quat()

        end_forward_kinematics = np.matmul(intermediate_rotation, start_forward_kinematics)

        end_position = end_forward_kinematics[:4, 3]
        end_orientation = Rotation.from_matrix(end_forward_kinematics[:3, :3]).as_quat()

        if self.index == 2:
            self.publish_pose_target_pose(end_position, end_orientation)


    def transformation_simple_robot(self):
        # Define the desired rotation here
        theta = - pi / 1.8
        # Split theta
        theta_fraction = self.phi * theta

        self.publish_angle(theta_fraction)

        # Define the matrix
        intermediate_rotation = np.array([
            [cos(theta_fraction), -sin(theta_fraction), 0, 0],
            [sin(theta_fraction), cos(theta_fraction), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ])

        self.simple_robot_joint_angles = np.zeros(len(self.simple_robot_Slist[0]))
        # self.simple_robot_joint_angles[1] = np.pi / 2

        start_forward_kinematics_simple = FKinBody(self.simple_robot_M, self.simple_robot_Blist, self.simple_robot_joint_angles)

        start_position_simple = start_forward_kinematics_simple[:4 ,3]
        start_orientation = Rotation.from_matrix(start_forward_kinematics_simple[:3, :3]).as_quat()

        end_forward_kinematics_simple = np.matmul(intermediate_rotation, start_forward_kinematics_simple)

        end_position_simple = end_forward_kinematics_simple[:4, 3]
        end_orientation_simple = Rotation.from_matrix(end_forward_kinematics_simple[:3, :3]).as_quat()


        self.publish_pose_simple_robot(end_position_simple, end_orientation_simple)
        
        self.construct_path(end_position_simple)

    def construct_path(self, vector):
        # RViz
        pose = PoseStamped()
        pose.header.frame_id = "map"
        pose.header.stamp = self.get_clock().now().to_msg()
        pose.pose.position.x = vector[0]
        pose.pose.position.y = vector[1]
        pose.pose.position.z = vector[2]
        pose.pose.orientation.w = 1.0

        # if the path is completed clear it
        if self.phi == 0:
            self.path.poses.clear()
        else:
            self.path.poses.append(pose)

        # publish the path
        self.path_publisher_.publish(self.path)

        # call the function publishing the marker
        self.publish_marker(pose.pose)

    def publish_marker(self, current_pose):
        # RViz
        marker = Marker()
        marker.header.frame_id = "map"
        marker.header.stamp = self.get_clock().now().to_msg()
        marker.type = Marker.SPHERE
        marker.action = Marker.ADD
        marker.pose = current_pose
        marker.scale.x = 0.05
        marker.scale.y = 0.05
        marker.scale.z = 0.05
        marker.color.a = 1.0
        marker.color.r = 1.0
        marker.color.g = 0.0
        marker.color.b = 0.0

        # publish the marker
        self.point_publisher_.publish(marker)


    def publish_angle(self, angle):
        angle_msg = Float64()
        angle_msg.data = angle * 180 / pi
        self.angle_publisher_.publish(angle_msg)


    def publish_pose_simple_robot(self, vector, quaternion):
        pose_msg_sr = Pose()
        # pose_msg.header.frame_id = "map"
        # pose_msg.header.stamp = self.get_clock().now().to_msg()
        pose_msg_sr.position.x = vector[0]
        pose_msg_sr.position.y = vector[1]
        pose_msg_sr.position.z = vector[2]
        pose_msg_sr.orientation.x = quaternion[0]
        pose_msg_sr.orientation.y = quaternion[1]
        pose_msg_sr.orientation.z = quaternion[2]
        pose_msg_sr.orientation.w = quaternion[3]
        self.pose_publisher_simple_robot_.publish(pose_msg_sr)

    def publish_pose_target_pose(self, vector, quaternion):    
        # Create a Pose message for the target pose
        pose_msg = Pose()
        pose_msg.position.x = vector[0]
        pose_msg.position.y = vector[1]
        pose_msg.position.z = vector[2]
        pose_msg.orientation.x = quaternion[0]
        pose_msg.orientation.y = quaternion[1]
        pose_msg.orientation.z = quaternion[2]
        pose_msg.orientation.w = quaternion[3]
        self.pose_publisher_kuka_robot_.publish(pose_msg)
        # if self.phi == 0.0:
        #     time.sleep(2)


def main(args = None):
    rclpy.init(args = args)
    node = Forward_Kinematics_Node_1Dof()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

