import os
import re

import rclpy
from rclpy.node import Node
from geometry_msgs.msg import Pose
from sensor_msgs.msg import JointState
from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from modern_robotics import IKinBody, RpToTrans, FKinBody
import numpy as np
from ament_index_python.packages import get_package_share_directory
from mr_urdf_loader import loadURDF
from scipy.spatial.transform import Rotation
from rclpy.clock import Clock

class InverseKinematicsSimpleRobot(Node):
    def __init__(self):
        super().__init__('inverse_kinematics_simple_robot')
        self.subscription = self.create_subscription(
            Pose,
            '/simple_robot/target_pose',
            self.pose_callback,
            10)
        self.publisher = self.create_publisher(JointState, '/joint_states', 10)

        urdf_file_path = os.path.join(
            get_package_share_directory('simple_robot_1dof'), 'urdf', 'simple_robot.urdf')
        self.get_logger().info(f'Loading URDF from: {urdf_file_path}')
        self.simple_robot = loadURDF(urdf_file_path)

        self.M = self.simple_robot["M"]
        self.Slist = self.simple_robot["Slist"]
        self.Mlist = self.simple_robot["Mlist"]
        self.Glist = self.simple_robot["Glist"]
        self.Blist = self.simple_robot["Blist"]

        self.joint_names = [f'joint{i+1}' for i in range(len(self.Slist[0]))]

        self.joint_angles = np.zeros(len(self.Slist[0]))

        T = FKinBody(self.M, self.Blist, self.joint_angles)
        rotation_matrix = T[:3, :3]
        quat = Rotation.from_matrix(rotation_matrix).as_quat()
        position = T[:3, 3]
        self.get_logger().info('T: "%s":' %T)

        self.get_logger().info(f'Calculated Forward Kinematics Quaternion: {quat}')
        self.get_logger().info(f'Calculated Forward Kinematics Position: {position}')

    def pose_callback(self, pose: Pose):
        quat = [pose.orientation.x, pose.orientation.y, pose.orientation.z, pose.orientation.w]
        r = Rotation.from_quat(quat)
        rotation_matrix = r.as_matrix()
        T_sd = RpToTrans(rotation_matrix, np.array([pose.position.x, pose.position.y, pose.position.z]))
        eomg = 1e-4  # Fehlergrenze für Winkel
        ev = 1e-4  # Fehlergrenze für Translation

        joint_angles, success = IKinBody(self.Blist, self.M, T_sd, self.joint_angles, eomg, ev)
       

        if success:
            # self.joint_angles = joint_angles


            joint_state_msg = JointState()
            joint_state_msg.header.stamp = Clock().now().to_msg()
            joint_state_msg.name = self.joint_names
            joint_state_msg.position = joint_angles.tolist()

            # joint_trajectory_msg = JointTrajectory()
            # joint_trajectory_msg.joint_names = self.joint_names
            # point = JointTrajectoryPoint()
            # point.positions = joint_angles.tolist()
            # point.time_from_start.sec = 1 # Example time, adjust as needed
            # point.time_from_start.nanosec = 450000000 # Balint fragen, ob das verbessert werden kann
            # joint_trajectory_msg.points.append(point)

            # self.publisher.publish(joint_trajectory_msg)
            self.publisher.publish(joint_state_msg)
        else:
            self.get_logger().error("Inverse Kinematics failed to find a solution.")


def main(args=None):
    rclpy.init(args=args)
    node = InverseKinematicsSimpleRobot()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()